#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#include <sys/socket.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/time.h>

#include "tunnelit-common.h"
#include "tunnelit-helper.h"
#include "ethertun.h"
#include "net.h"

typedef enum {
	TUNNELIT_OPMODE_CONNECT = 0,
	TUNNELIT_OPMODE_LISTEN = 1,
	TUNNELIT_OPMODE_TCP = 0,
	TUNNELIT_OPMODE_UDP = 2
} opmode_t;

static fd_set *copy_sock_to_tun(int srcsock, int tun, int socks[TUNNELIT_MAX_CHILDREN], int highestchildidx) {
	static fd_set retval;
	static char buf[16384], *buf_ptr;
	ssize_t recv_ret, write_ret, send_ret, tosend;
	int sock;
	int idx;

	FD_ZERO(&retval);

	recv_ret = recv(srcsock, buf, sizeof(buf), 0);
	if (recv_ret <= 0) {
		DEBUG_LOG("debug", "Got EOF from Socket (%i), closing it.", srcsock);
		FD_SET(srcsock, &retval);
		return(&retval);
	}

	tosend = recv_ret;
	buf_ptr = buf;
	while (tosend) {
		write_ret = write(tun, buf_ptr, tosend);
		if (write_ret <= 0) {
			DEBUG_LOG("debug", "Got EOF from TUN device (%i), odd.", tun);
			return(&retval);
		}

		tosend -= write_ret;
		buf_ptr += write_ret;
	}

	for (idx = 0; idx < (sizeof(socks) / sizeof(socks[0])); idx++) {
		if (idx > highestchildidx) {
			break;
		}

		if (socks[idx] < 0) {
			continue;
		}

		if (socks[idx] == srcsock) {
			continue;
		}

		tosend = recv_ret;
		buf_ptr = buf;
		sock = socks[idx];

		while (tosend) {
			send_ret = send(sock, buf_ptr, tosend, 0);
			if (send_ret <= 0) {
				DEBUG_LOG("debug", "Got EOF from Socket (%i), closing it.", sock);
				FD_SET(sock, &retval);
				continue;
			}

			tosend -= send_ret;
			buf_ptr += send_ret;
		}
	}

	return(&retval);
}

static fd_set *copy_tun_to_socks(int tun, int socks[TUNNELIT_MAX_CHILDREN], int highestchildidx) {
	static fd_set retval;
	static char buf[16384], *buf_ptr;
	ssize_t read_ret, send_ret, tosend;
	int sock;
	int idx;

	FD_ZERO(&retval);

	read_ret = read(tun, buf, sizeof(buf));
	if (read_ret <= 0) {
		DEBUG_LOG("debug", "Got EOF from TUN (%i) !?", tun);
		return(&retval);
	}

	for (idx = 0; idx < (sizeof(socks) / sizeof(socks[0])); idx++) {
		if (idx > highestchildidx) {
			break;
		}

		if (socks[idx] < 0) {
			continue;
		}

		tosend = read_ret;
		buf_ptr = buf;
		sock = socks[idx];

		while (tosend) {
			send_ret = send(sock, buf_ptr, tosend, 0);
			if (send_ret <= 0) {
				DEBUG_LOG("debug", "Got EOF from Socket (%i), closing it.", sock);
				FD_SET(sock, &retval);
				continue;
			}

			tosend -= send_ret;
			buf_ptr += send_ret;
		}
	}

	return(&retval);
}

static int net_connect_tcp_persistent(const char *host, int port) {
	int newfd;

	newfd = -1;
	while (newfd < 0) {
		newfd = net_connect_tcp(host, port);
		if (newfd >= 0) {
			break;
		}

		PERROR("net_connect_tcp");
		sleep(5);
	}

	return(newfd);
}

static void print_help(FILE *chan) {
	fprintf(chan, "Usage: tunnelit-p2p [-Vhlu] [-p port] [-H host] [-d devname]\n");
	fprintf(chan, "   -V            Print version and exit\n");
	fprintf(chan, "   -h            Print this help and exit\n");
	fprintf(chan, "   -l            Listen (default is connect)\n");
	fprintf(chan, "   -u            UDP mode (not implemented)\n");
	fprintf(chan, "   -H <host>     Connect to host, required if \"-l\" not specified\n");
	fprintf(chan, "   -d <devname>  TUN/TAP device name\n");
	return;
}

int main(int argc, char **argv) {
	opmode_t opmode = TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP;
	fd_set readfds, errfds, *copyfailfds;
	char *host = NULL, *devname = NULL;
	char ch;
	int tunfd, masterfd, newfd, childfds[TUNNELIT_MAX_CHILDREN], maxfd;
	int port = 656;
	int select_ret;
	int idx;

	while ((ch = getopt(argc, argv, "Vhlup:H:d:")) != -1) {
		switch (ch) {
			case 'V':
				printf("TunnelIt Point-to-Point version %s-%s", PACKAGE_VERSION,
#ifdef DEBUG
				       "debug"
#else
				       "rel"
#endif
				);
				return(0);
				break;
			case 'h':
				print_help(stdout);
				return(0);
				break;
			case 'l':
				opmode |= TUNNELIT_OPMODE_LISTEN;
				break;	
			case 'u':
				opmode |= TUNNELIT_OPMODE_UDP;
				break;
			case 'p':
				port = atoi(optarg);
				break;
			case 'H':
				host = optarg;
				break;
			case 'd':
				devname = strdup(optarg);
				break;
			case '?':
			case ':':
				print_help(stderr);
				return(1);
				break;
		}
	}

	if ((opmode & TUNNELIT_OPMODE_LISTEN) == TUNNELIT_OPMODE_CONNECT) {
		if (!host) {
			ERROR_LOG("Must specify hostname (-H)");
			return(1);
		}
	}

	tunfd = tun_alloc(devname);
	if (tunfd < 0) {
		PERROR("tun_alloc");
		ERROR_LOG("Unable to create tun/tap device");
		return(1);
	}

	for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
		childfds[idx] = -1;
	}

	masterfd = -1;
	switch ((int) opmode) {
		case TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP:
			newfd = net_connect_tcp_persistent(host, port);

			childfds[0] = newfd;

			break;
		case TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_UDP:
			ERROR_LOG("not implemented");
			return(1);
			break;
		case TUNNELIT_OPMODE_LISTEN | TUNNELIT_OPMODE_UDP:
			ERROR_LOG("not implemented");
			return(1);
			break;
		case TUNNELIT_OPMODE_LISTEN | TUNNELIT_OPMODE_TCP:
			masterfd = -1;
			while (masterfd < 0) {
				masterfd = net_listen_tcp(port);
				if (masterfd >= 0) {
					break;
				}

				PERROR("net_listen_tcp");
				sleep(5);
			}

			break;
	}

	while (1) {
		FD_ZERO(&readfds);
		FD_ZERO(&errfds);

		maxfd = 0;
		for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
			if (childfds[idx] < 0) {
				continue;
			}

			FD_SET(childfds[idx], &readfds);
			FD_SET(childfds[idx], &errfds);
			if (childfds[idx] > maxfd) {
				maxfd = childfds[idx];
			}
		}
		if (tunfd > maxfd) {
			maxfd = tunfd;
		}
		FD_SET(tunfd, &readfds);
		FD_SET(tunfd, &errfds);

		if (masterfd >= 0) {
			if (masterfd > maxfd) {
				maxfd = masterfd;
			}
			FD_SET(masterfd, &readfds);
			FD_SET(masterfd, &errfds);
		}

		select_ret = select(maxfd + 1, &readfds, NULL, &errfds, NULL);
		if (select_ret < 0) {
			sleep(5);
			continue;
		}

		if (select_ret) {
			if (FD_ISSET(tunfd, &errfds)) {
				select_ret--;
				ERROR_LOG("TUN device (%i) had an error, aborting.", tunfd);
				return(1);
			}

			if (FD_ISSET(tunfd, &readfds)) {
				select_ret--;
				DEBUG_LOG("debug", "Copying from TUN device (%i) to Socket (%i)", tunfd, childfds[0]);
				copyfailfds = copy_tun_to_socks(tunfd, childfds, TUNNELIT_MAX_CHILDREN);
				for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
					if (childfds[idx] < 0) {
						continue;
					}

					if (FD_ISSET(childfds[idx], copyfailfds)) {
						DEBUG_LOG("debug", "Child Socket (%i) had an error, dropping it", childfds[idx]);
						close(childfds[idx]);
						childfds[idx] = -1;

						if (idx == 0 && opmode == (TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP)) {
							DEBUG_LOG("debug", "Attempting to reconnect to %s:%i...", host, port);
							childfds[0] = net_connect_tcp_persistent(host, port);
						}
					}
				}
			}
		}

		if (select_ret && masterfd >= 0) {
			if (FD_ISSET(masterfd, &errfds)) {
				select_ret--;
				ERROR_LOG("Listening Socket (%i) had an error, aborting.", masterfd);
				return(1);
			}

			if (FD_ISSET(masterfd, &readfds)) {
				select_ret--;
				DEBUG_LOG("debug", "Accepting new connection (%i)", masterfd);
				newfd = -1;
				while (newfd < 0) {
					newfd = accept(masterfd, NULL, NULL);
					if (newfd >= 0) {
						break;
					}

					PERROR("accept");
					sleep(5);
				}

				for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
					if (childfds[idx] < 0) {
						childfds[idx] = newfd;
						break;
					}
				}

			}
		}

		idx = 0;
		while (select_ret) {
			for (; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
				if (childfds[idx] < 0) {
					continue;
				}

				if (FD_ISSET(childfds[idx], &errfds)) {
					select_ret--;
					DEBUG_LOG("debug", "Child Socket (%i) had an error, dropping it", childfds[idx]);
					close(childfds[idx]);
					childfds[idx] = -1;

					if (idx == 0 && opmode == (TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP)) {
						DEBUG_LOG("debug", "Attempting to reconnect to %s:%i...", host, port);
						childfds[0] = net_connect_tcp_persistent(host, port);
					}
					break;
				}

				if (FD_ISSET(childfds[idx], &readfds)) {
					select_ret--;
					DEBUG_LOG("debug", "Copying from Socket (%i) to TUN device (%i)", childfds[idx], tunfd);
					copyfailfds = copy_sock_to_tun(childfds[idx], tunfd, childfds, TUNNELIT_MAX_CHILDREN);

					for (idx = 0; idx < (sizeof(childfds) / sizeof(childfds[0])); idx++) {
						if (childfds[idx] < 0) {
							continue;
						}

						if (FD_ISSET(childfds[idx], copyfailfds)) {
							DEBUG_LOG("debug", "Child Socket (%i) had an error, dropping it", childfds[idx]);
							close(childfds[idx]);
							childfds[idx] = -1;

							if (idx == 0 && opmode == (TUNNELIT_OPMODE_CONNECT | TUNNELIT_OPMODE_TCP)) {
								DEBUG_LOG("debug", "Attempting to reconnect to %s:%i...", host, port);
								childfds[0] = net_connect_tcp_persistent(host, port);
							}
						}
					}

					break;
				}
			}
		}
	}

	close(tunfd);

	return(0);
}
